"""
Collectors are used to collect multiple candidates generated by generators. 
This is a task that can be parallelized over the number of generated candidates, 
and this is done using parallelization from Ray. 
"""
from agox.collectors.standard import StandardCollector
from agox.collectors.ray_collector import ParallelCollector
from agox.collectors.parallel_tempering import ParallelTemperingCollector

__all__ = ['StandardCollector', 'ParallelCollector', 'ParallelTemperingCollector']
