from agox.test.run_tests.run_utils import agox_test_run
import pytest

@pytest.mark.skip(reason='Not reproducible on CI at the moment.')
def test_gofee_cluster_symmetry(tmp_path, cmd_options, ray_fix):
    mode = 'gofee_cluster_symmetry' # This determines the script that is imported.
    agox_test_run(mode, tmp_path, cmd_options, full_compare=False)
